# -*- coding: utf-8 -*-

"""Collection of the core utilities."""

from __future__ import annotations

__all__ = ['get_conda_cmd_path', 'is_conda_available', 'get_conda_info']

import itertools
import json
import os
import sys
import typing

from . import exceptions
from . import launch
from . import solvers

if typing.TYPE_CHECKING:
    from . import types as conda_types


CONDA_PATH: str
if os.name == 'nt':
    CONDA_PATH = os.path.join('Scripts', 'conda-script.py')
else:
    CONDA_PATH = os.path.join('bin', 'conda')


def get_conda_cmd_path(*prefixes: str) -> typing.Optional[str]:
    """Check if conda is found on path."""
    commands: typing.List[str] = []

    conda_exe: str = os.environ.get('CONDA_EXE', '')
    if conda_exe:
        commands.append(conda_exe)

    prefix: str = os.path.abspath(sys.prefix)
    for prefix in itertools.chain(prefixes, [os.path.dirname(os.path.dirname(prefix)), prefix]):
        commands.append(os.path.join(prefix, CONDA_PATH))

    commands.append('conda')

    command: str
    for command in commands:
        stdout: str
        stderr: str
        error: bool
        stdout, stderr, error = launch.run_process(cmd_list=[command, '--version'])

        if (not error) and any(item.startswith('conda ') for item in (stdout, stderr)):
            return command

    return None


def is_conda_available() -> bool:
    """Check if conda is available in path."""
    return get_conda_cmd_path() is not None


def get_conda_info(allow_fixes: bool = False) -> typing.Optional['conda_types.CondaInfoOutput']:
    """Return conda info as a dictionary."""
    conda_cmd: typing.Optional[str] = get_conda_cmd_path()
    if conda_cmd is None:
        return None

    while True:
        out: str = launch.run_process(cmd_list=[conda_cmd, 'info', '--json']).stdout

        result: typing.Union['conda_types.CondaInfoOutput', 'conda_types.CondaErrorOutput']
        try:
            result = json.loads(out)
        except (TypeError, ValueError):
            return None

        if 'error' not in result:
            return typing.cast('conda_types.CondaInfoOutput', result)

        error: 'conda_types.CondaErrorOutput' = typing.cast('conda_types.CondaErrorOutput', result)
        if allow_fixes and (solvers.POOL.solve(error=error) is not None):
            continue
        raise exceptions.CondaError(error=error)



# -*- coding: utf-8 -*-

"""Collection of utilities for importing python entities."""

from __future__ import annotations

__all__ = ['import_callable', 'import_class']

import typing
from . import exceptions

if typing.TYPE_CHECKING:
    import typing_extensions


TTypeInv = typing.TypeVar('TTypeInv', bound=type)


def _import_anything(name: typing.Optional[str]) -> typing.Any:
    """Locate any item by its import path."""
    if name is None:
        raise exceptions.ValidationError('value is required')

    if name.startswith('::') and ('.' not in name):
        name = name[2:]
    else:
        raise exceptions.ValidationError('invalid identifier')

    from . import bundle  # pylint: disable=cyclic-import,import-outside-toplevel
    result = getattr(bundle, name, None)

    if result is None:
        raise exceptions.ValidationError('invalid identifier')

    return result


@typing.overload
def import_callable(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[False]' = False,
) -> typing.Callable[..., typing.Any]:
    ...


@typing.overload
def import_callable(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[True]',
) -> typing.Optional[typing.Callable[..., typing.Any]]:
    ...


def import_callable(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[True, False]' = False,
) -> typing.Optional[typing.Callable[..., typing.Any]]:
    """
    Locate a callable object by its import path.

    :param name: Path to callable to import.

                 E.g. :code:`'json.loads'`
    :param allow_none: Allow `name` to be :code:`None`.

                       If allowed and `name` is None - :code:`None` will be returned.
    """
    if (name is None) and allow_none:
        return None

    result: typing.Any = _import_anything(name)

    if not callable(result):
        raise exceptions.ValidationError('is not a function')

    return result


@typing.overload
def import_class(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[False]' = False,
        inherited_from: typing.Union[TTypeInv, typing.Sequence[TTypeInv]] = (),
) -> TTypeInv:
    ...


@typing.overload
def import_class(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[True]',
        inherited_from: typing.Union[TTypeInv, typing.Sequence[TTypeInv]] = (),
) -> typing.Optional[TTypeInv]:
    ...


def import_class(
        name: typing.Optional[str],
        *,
        allow_none: 'typing_extensions.Literal[True, False]' = False,
        inherited_from: typing.Union[TTypeInv, typing.Sequence[TTypeInv]] = (),
) -> typing.Optional[typing.Type]:
    """
    Locate a class object by its import path

    :param name: Path to callable to import.

                 E.g. :code:`'json.JSONEncoder'`
    :param allow_none: Allow `name` to be :code:`None`.

                       If allowed and `name` is None - :code:`None` will be returned.
    :param inherited_from: Validate that target class is inherited from one of optional bases
    """
    if (name is None) and allow_none:
        return None

    result: typing.Any = _import_anything(name)

    if not isinstance(result, type):
        raise exceptions.ValidationError('is not a class')

    parents: typing.Tuple[type, ...]
    if isinstance(inherited_from, type):
        parents = (inherited_from,)
    else:
        parents = tuple(typing.cast(typing.Sequence[TTypeInv], inherited_from))

    if inherited_from and not issubclass(result, parents):
        raise exceptions.ValidationError('this class can not be used')

    return result



# -*- coding: utf-8 -*-

"""Collection of basic utilities."""

from __future__ import annotations

__all__ = ['coalesce']

import typing

if typing.TYPE_CHECKING:
    import typing_extensions


TInv = typing.TypeVar('TInv', bound=object)


@typing.overload
def coalesce(*args: typing.Optional[TInv], allow_none: 'typing_extensions.Literal[False]' = False) -> TInv:
    """Find first not-:code:`None` value."""


@typing.overload
def coalesce(*args: typing.Optional[TInv], allow_none: 'typing_extensions.Literal[True]') -> typing.Optional[TInv]:
    """Find first not-:code:`None` value."""


def coalesce(
        *args: typing.Optional[TInv],
        allow_none: 'typing_extensions.Literal[True, False]' = False,
) -> typing.Optional[TInv]:
    """
    Find first not-:code:`None` value.

    :param allow_none: Allow :code:`None` to be the result if all values are :code:`None`.
    """
    arg: typing.Optional[TInv]
    for arg in args:
        if arg is not None:
            return arg
    if allow_none:
        return None
    raise ValueError('not-null not found')
